{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE TupleSections #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE MagicHash #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE ViewPatterns #-}
module Engine.Experiment
( ExperimentResult(..)
, runExperiment
, remapExperimentResult
) where

import Bottom
import Class
import Control.Applicative
import Control.Concurrent
import Control.Exception
import Control.Monad
import Control.Monad.Instances
import Control.Monad.Reader (ask)
import CoreSyn
import CoreUtils
import DataCon
import Data.Map (Map)
import Data.Maybe
import Data.Monoid
import Engine.EngineHValue
import Engine.EngineM
import Engine.Eval
import Engine.ExperimentResult
import Engine.IrulanRTError
import FastString
import Format
import GHC (HValue)
import GHC.Prim
import HscTypes
import Id
import InstEnv
import MkCore
import MonadUtils (liftIO)
import Plan
import PrelNames (showClassName)
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Traversable as T
import qualified UniqSet
import System.IO.Unsafe
import System.Time
import TcRnMonad
import TyCon
import Type
import TysPrim (anyPrimTy)
import TysWiredIn (unitDataConId, unitTy)
import Unsafe.Coerce
import Util
import Util.GHC
import Var

type InstTypes = [Either TyVar Type]
	-- Right ty	=> Instantiate with this type
	-- Left tv 	=> Instantiate with any type of this tyvar's kind

type InstMatch = (Instance, InstTypes)

instance Format ExperimentResult where
  format (Redundant exps) = "# == " ++ format exps
  format Ok              = "."
  format (Errors exc)    = "! " ++ show exc
  format (Bottoms bid)   = "?" ++ format bid
  format (Property p)
    | p                  = "True"
    | otherwise          = "False"
  format (Show s)        = s

runExperiment :: Experiment -> EngineM RanExperiment
runExperiment experiment = profileFunction "runExperiment" $ do
  mExistingResult <- lookupExperimentResult experiment
  
  case mExistingResult of
    Just existingResult -> do
      CachedResult existingResult <$> checkExperimentCache experiment
    Nothing -> do
      env <- ask

      val <- (if useShowInstances env && (not . experimentIsProperty $ experiment) 
                then mkExperimentShowValue 
                else mkExperimentValue) experiment 

      resultVar <- liftIO $ newEmptyMVar
      evalThreadId <- liftIO $ forkIO (evalWithSeq val >>= ((putMVar resultVar) $!))

      evalResult <- profileFunction "runExperiment:watchMemoryAndTime" (liftIO $ watchMemoryAndTime env resultVar)

      liftIO (killThread evalThreadId)
      liftIO getAllocations >>= \allocs -> tellLog ("Allocations: " ++ show allocs)

      let simpleCase result hv = storeExperimentResult experiment result >> return (result, hv)

      (!ranExp, !mEHVal) <- case evalResult of
        EvalException (fromException -> Just (Bottom i)) -> simpleCase (Bottoms i) Nothing
        EvalException e  -> simpleCase (Errors e) Nothing
        EvalOk val -> do
          if experimentIsProperty experiment
            then simpleCase (Property (unsafeCoerce# (toHValue val))) (Just val)
            else do
              if experimentHasBottom experiment
                then simpleCase Ok (Just val)
                else (maybe (Ok, Just val)) ((,Just val) . Redundant) <$> storeExperimentHValue experiment val
        EvalString str -> return $ (Show str, Nothing)
          
      return $! NewResult ranExp mEHVal

watchMemoryAndTime :: EngineEnv -> MVar EvalResult -> IO EvalResult
watchMemoryAndTime env er = do
  startTime <- getClockTime
  let endTime = noTimeDiff { tdPicosec = 1000000 * (fromIntegral (maxExpressionTime env)) } `addToClockTime` startTime 
  baseAllocs <- getAllocations
  let endAllocs = baseAllocs + maxExpressionAllocs env

  let loop = do
              newAllocs <- getAllocations
              newTime   <- getClockTime
              mVal      <- tryTakeMVar er

              let go
                    | Just res  <- mVal
                        = return res
                    | newAllocs > endAllocs 
                        = return (EvalException $! SomeException $! ErrorCall "<allocated too much memory!>")
                    | newTime > endTime
                        = return (EvalException $! SomeException $! ErrorCall "<ran out of time!>")
                    | otherwise = yield >> loop
              go
    
  loop

getOrMkExperimentValue :: Experiment -> EngineM CompiledExpression
getOrMkExperimentValue exp
  | not (experimentHasBottom exp) = profileFunction "getOrMkExperimentValue" $ (checkExperimentCache exp >>= maybe (mkExperimentValue exp)
                                                                                                                   (return . liftM2 CE toHValue toType))
  | otherwise = profileFunction "getOrMkExperimentValue" $ mkExperimentValue exp

mkExperimentShowValue :: Experiment -> EngineM CompiledExpression
mkExperimentShowValue !experiment = do
  ce@(CE hv ty) <- mkExperimentValue experiment

  mSF <- mkShowFunction ty
  case mSF of
    Nothing -> return $! ce
    Just sf -> return $! CEShow (sf hv)

mkShowFunction :: Type -> EngineM (Maybe (HValue -> String))
mkShowFunction ty = withSession $ \hscEnv -> do
  Just (AClass showClass) <- liftIO $ lookupTypeHscEnv hscEnv showClassName
  let [_, showId, _] = classMethods showClass
  eps <- liftIO . hscEPS $ hscEnv
  let (home_instances, _) = hptInstances hscEnv (const True)
  let homeEnv = extendInstEnvList emptyInstEnv home_instances
  let inst_envs = (eps_inst_env eps, homeEnv)

  let freeVars = tyVarsOfType ty
  let ty' = substTy (mkTopTvSubst [ (tv, unitTy) | tv <- UniqSet.uniqSetToList freeVars ] ) ty

  mRes <- getDictionary inst_envs showClass [ty']

  case mRes of
    Nothing -> return Nothing
    Just res -> do
      showFn  <- getHValue showId
      return $ Just $ ((unsafeCoerce# showFn) res)

getDictionary :: (InstEnv, InstEnv) -> Class -> [Type] -> EngineM (Maybe HValue)
getDictionary inst_envs cls tys = do
  let (matches, _) = lookupInstEnv inst_envs cls tys

  let getDictIdAndType = do
        [m] <- return matches
        let (inst, instETys) :: InstMatch = unsafeCoerce m
        let dictId = instanceDFunId inst
        instTys <- mapM (either (const Nothing) (Just)) instETys
        let dictIdType = applyTys (idType dictId) instTys
        -- () <- mTrace ("mESV: " ++ show dictId ++ " :: " ++ show dictIdType)
        return (dictId, dictIdType)

  case getDictIdAndType of
    Nothing -> return Nothing
    Just (dictId, dictIdType) -> do
      let (argTys, _) = splitFunTys dictIdType

      let processArg ty = 
            maybe (return Nothing) (uncurry $ getDictionary inst_envs) $ do 
              (tc, tas) <- splitTyConApp_maybe ty
              cls <- tyConClass_maybe tc
              return (cls, tas)

      mArgHValues <- mapM processArg argTys
      instVal <- getHValue dictId

      return $ foldM (\hv nxt -> (unsafeCoerce# hv) <$> nxt) instVal mArgHValues
  

mkExperimentValue :: Experiment -> EngineM CompiledExpression
mkExperimentValue !experiment = profileFunction "mkExperimentValue" $ do
  rootHVal <- mkExperimentRootValue
  childHVals <- catMaybes . F.toList <$> T.mapM mkArgumentValue (experimentArgs experiment)

  let hv = applyHValues rootHVal childHVals
  let ty = experimentType experiment

  return $! CE hv ty
  where
    mkExperimentRootValue :: EngineM HValue
    mkExperimentRootValue
      | ConstantRoot c <- experimentRoot experiment = return $! mkConstantValue c
      | IdRoot x <- experimentRoot experiment = getHValue x
      | DestructRoot { destructCaseOf 
                     , destructConstructor 
                     , destructVarIndex 
                     , destructResultType 
                     } <- experimentRoot experiment = do
        
      (CE hv _) <- getOrMkExperimentValue destructCaseOf
      return $ unsafeCoerce# (unsafePerformIO $ getConstructorArgument hv destructVarIndex)

    mkArgumentValue :: ExperimentArg -> EngineM (Maybe HValue)
    mkArgumentValue (ArgExperiment exp) = (\(CE hv _) -> Just hv) <$> getOrMkExperimentValue exp
    mkArgumentValue (ArgBottom _ bid)   = return $ Just (unsafeCoerce# (throw (Bottom bid)))
    mkArgumentValue (ArgType ty)        = return $ Nothing
    mkArgumentValue ArgUnconstrained    = return $ Just (unsafeCoerce# ())
    
    applyHValues :: HValue -> [HValue] -> HValue
    applyHValues v [] = v
    applyHValues v (x:xs) = applyHValues ((unsafeCoerce# v) x) xs

    mkConstantValue :: Constant -> HValue
    mkConstantValue (IntConstant c) = unsafeCoerce# c
    mkConstantValue (CharConstant c) = unsafeCoerce# c
    mkConstantValue (FloatConstant c) = unsafeCoerce# c
    mkConstantValue (DoubleConstant c) = unsafeCoerce# c
    mkConstantValue (IntegerConstant c _) = unsafeCoerce# c
