{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE ForeignFunctionInterface #-}
module FrontEnd.Stats
( StatsOutputter
, withStatsOutput
, outputStats
, SummaryData(..)
, setIterativeDeepeningDepth
) where

import Control.Applicative
import Control.Category
import Control.Exception
import Control.Monad
import Control.Monad.Trans
import Data.IORef
import Data.List
import Data.Map (Map)
import Data.Maybe
import Data.Ratio((%))
import Data.Time
import Engine
import Exception
import Foreign
import Foreign.C.String
import Foreign.C.Types
import GHC
import Plan
import Prelude hiding ((.), mod, id)
import qualified Data.Map as Map
import System.FilePath
import System.IO
import Trace.Hpc.Mix
import Trace.Hpc.Tix
import Trace.Hpc.Util
import Util.FCLabels
import Util.GHC
import Util.HpcReflect

data StatsOutputter = SO !Handle !(IORef SummaryData) !(Map String [Bool])
                    | SSummary   !(IORef SummaryOnlyData)
                    | NoStats

data SummaryData
  = SummaryData
  { ranExperimentSummary                 :: !ExperimentResultSummary
  , cachedExperimentSummary              :: !ExperimentResultSummary
  , firstTimestamp                       :: !UTCTime
  , lastCodeCoverage                     :: !Integer
  , lastTimestamp                        :: !UTCTime
  , experimentNum                        :: !Int
  , lastExperimentWritten                :: !Bool
  , lastIterativeDeepeningDepthCompleted :: !Int
  }

data SummaryOnlyData
  = SummaryOnlyData
  { soRanExperimentSummary      :: !ExperimentResultSummary
  , soCachedExperimentSummary   :: !ExperimentResultSummary
  , soExperimentNum             :: !Int
  }

data ExperimentResultSummary
  = ExperimentResultSummary
  { redundants       :: !Int
  , oks              :: !Int
  , errors           :: !Int
  , bottoms          :: !Int
  , propTrue         :: !Int
  , propFalse        :: !Int
  }

mkLabels [''SummaryData, ''SummaryOnlyData, ''ExperimentResultSummary]


emptySummaryData :: UTCTime -> SummaryData
emptySummaryData timeNow
  = SummaryData { ranExperimentSummary    = emptyExperimentResultSummary
                , cachedExperimentSummary = emptyExperimentResultSummary
                , firstTimestamp          = timeNow
                , lastCodeCoverage        = 0
                , lastTimestamp           = timeNow
                , experimentNum           = 0
                , lastExperimentWritten   = False
                , lastIterativeDeepeningDepthCompleted = -1
                }

emptySummaryOnlyData :: SummaryOnlyData
emptySummaryOnlyData 
  = SummaryOnlyData { soRanExperimentSummary    = emptyExperimentResultSummary
                    , soCachedExperimentSummary = emptyExperimentResultSummary
                    , soExperimentNum           = 0
                    }

emptyExperimentResultSummary :: ExperimentResultSummary
emptyExperimentResultSummary
  = ExperimentResultSummary { redundants = 0
                            , oks        = 0
                            , errors     = 0
                            , bottoms    = 0
                            , propTrue   = 0
                            , propFalse  = 0
                            }


withStatsOutput :: (MonadIO m, ExceptionMonad m) => Maybe FilePath -> Bool -> (StatsOutputter -> m a) -> m a
withStatsOutput Nothing _ f  = f NoStats
withStatsOutput (Just p) True f = gbracket setup shutdown f 
  where
    setup = liftIO $ do
      mixMap <- initMixMap

      timeNow <- getCurrentTime
      ref <- newIORef (emptySummaryData timeNow)
      hdl <- openFile p WriteMode

      return $! (SO hdl ref mixMap)

    shutdown (SO hdl ref mixMap) = liftIO $ do
      endStats <- readIORef ref

      unless (lastExperimentWritten endStats) $ do
        let timeStr = init $ show ((lastTimestamp endStats) `diffUTCTime` (firstTimestamp endStats))
        hPutStrLn hdl $ unwords [ show (experimentNum endStats - 1)
                                , timeStr
                                , show (lastCodeCoverage endStats)
                                ]

      hClose hdl
      hdl' <- openFile (p <.> "summary") WriteMode
      let maxPossible = maxTixsPossible mixMap

      let ccPercent :: Double
            | maxPossible == 0 = 0
            | otherwise        = fromRational $ (lastCodeCoverage endStats) % maxPossible

      let summarise lhs field = [ lhs
                                , show . field . ranExperimentSummary $ endStats
                                , show . field . cachedExperimentSummary $ endStats
                                ]

      hPutStr hdl' $ unlines $ map unwords
        [ "Redundant"      `summarise` redundants
        , "Ok"             `summarise` oks
        , "Errors"         `summarise` errors
        , "?Arg"           `summarise` bottoms
        , "Property-T"     `summarise` propTrue
        , "Property-F"     `summarise` propFalse
        , []
        , "CCGot"          `is` lastCodeCoverage endStats
        , "CCPossibleMax"  `is` maxPossible
        , "CCPercent"      `is` ccPercent
        , ["Total-Time"    , (init . show $ ((lastTimestamp endStats) `diffUTCTime` (firstTimestamp endStats)))]
        , "Total-Num-Exps" `is` experimentNum endStats
        ]
      hClose hdl'
withStatsOutput (Just p) False f = gbracket setup shutdown f
  where
    setup = liftIO $ do
      ref <- newIORef emptySummaryOnlyData
      return $! SSummary ref
    
    shutdown (SSummary ref) = liftIO $ do
      endStatsOnly <- readIORef ref

      hdl <- openFile (p <.> "summary") WriteMode

      let summarise lhs field = [ lhs
                                , show . field . soRanExperimentSummary $ endStatsOnly
                                , show . field . soCachedExperimentSummary $ endStatsOnly
                                ]

      hPutStr hdl $ unlines $ map unwords
        [ "Redundant"      `summarise` redundants
        , "Ok"             `summarise` oks
        , "Errors"         `summarise` errors
        , "?Arg"           `summarise` bottoms
        , "Property-T"     `summarise` propTrue
        , "Property-F"     `summarise` propFalse
        , []
        , "Total-Num-Exps" `is` soExperimentNum endStatsOnly
        ]

      hClose hdl


is :: (Show a) => String -> a -> [String]
is x y = [x, show y]


outputStats :: StatsOutputter -> Experiment -> RanExperiment -> UTCTime -> IO ()
outputStats NoStats _ _ _                    = return ()
outputStats (SO hdl ref mixs) exp ranExp expStartTime = do
  tix <- getTotalExpressionWrites mixs <$> examineTix
  sd  <- readIORef ref

  let writeCondition = lastCodeCoverage sd < tix
  let startTime = firstTimestamp sd

  when writeCondition $ do
    let timeStr = init $ show (expStartTime `diffUTCTime` startTime)
    hPutStrLn hdl $ unwords [show (experimentNum sd), timeStr, show tix]

  writeIORef ref $! mod (resultElem . newOrCached) (+1)
                  . set lLastCodeCoverage tix
                  . set lLastTimestamp expStartTime
                  . mod lExperimentNum (+1)
                  . set lLastExperimentWritten writeCondition
                  $ sd

  where
    newOrCached = case ranExp of
                NewResult _ _    -> lRanExperimentSummary
                CachedResult _ _ -> lCachedExperimentSummary
    
    resultElem = case (getExperimentResult ranExp) of
                    Redundant _    -> lRedundants
                    Ok             -> lOks
                    Show _         -> lOks
                    Errors _       -> lErrors
                    Bottoms _      -> lBottoms
                    Property True  -> lPropTrue
                    Property False -> lPropFalse
outputStats (SSummary ref) exp ranExp _ = do
  sd <- readIORef ref

  writeIORef ref $! mod (resultElem . newOrCached) (+1)
                  . mod lSoExperimentNum (+1)
                  $ sd
  where
    newOrCached = case ranExp of
                NewResult _ _    -> lSoRanExperimentSummary
                CachedResult _ _ -> lSoCachedExperimentSummary
    
    resultElem = case (getExperimentResult ranExp) of
                    Redundant _    -> lRedundants
                    Ok             -> lOks
                    Show _         -> lOks
                    Errors _       -> lErrors
                    Bottoms _      -> lBottoms
                    Property True  -> lPropTrue
                    Property False -> lPropFalse

setIterativeDeepeningDepth :: StatsOutputter -> Int -> IO ()
setIterativeDeepeningDepth NoStats _ = return ()
setIterativeDeepeningDepth (SO _ ref _) v = modifyIORef ref (\sd -> sd { lastIterativeDeepeningDepthCompleted = v })
