{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE NamedFieldPuns #-}
module TestSuite.TestCase where

import Bottom
import Control.Arrow
import DataCon
import Data.IntMap (IntMap)
import Data.Map (Map)
import Data.Maybe
import Data.Monoid
import Data.Sequence (Seq)
import Data.Set (Set)
import Data.String
import Data.Text (Text)
import Format
import Outputable
import qualified Data.Foldable as F
import qualified Data.IntMap as IntMap
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import qualified Data.Set as Set
import qualified Data.Text as T
import qualified Engine.ExperimentResult as E
import qualified Plan.Experiment as E hiding (ExperimentResult(..)) 
import Util

data TestCase 
  = TestCase { testExperiment :: !TestExperiment
             , testResult     :: !TestResult
             }
  deriving (Eq, Ord)

data TestResult
  = Errors   !Text
  | Property !Bool
  | Show     !Text
  deriving (Eq, Ord)

data TestExperiment
  = TestExperiment { teRoot :: !TestExperimentRoot
                   , teArgs :: !(Seq TestExperimentArg)
                   }
  deriving (Eq, Ord)

data TestExperimentRoot
  = TEIdRoot !Text
  | TEDestructRoot
  { teDestructCaseOf             :: !TestExperiment
  , teDestructConstructor        :: !Text
  , teDestructConstructorArity   :: !Int
  , teDestructConstructorIsInfix :: !Bool
  , teDestructVarIndex           :: !Int
  }
  | TEConstantRoot TestExperimentConstant
  deriving (Eq, Ord)

data TestExperimentConstant
  = TECInt     !Int
  | TECChar    !Char
  | TECFloat   !Float
  | TECDouble  !Double
  | TECInteger !Integer
  deriving (Eq, Ord)

data TestExperimentArg
  = TEArgExperiment !TestExperiment
  | TEArgUnconstrained
  | TEArgBottom !BottomId
  deriving (Eq, Ord)


data TestComparison
  = TC { moreGeneral :: !TestSuiteName
       , general     :: !TestResult
       , specific    :: !(Set TestCase)
       }

data TestSuiteName = A | B
  deriving Show

otherName :: TestSuiteName -> TestSuiteName
otherName A = B
otherName B = A

data DefinedComparison
  = LessDefined
  | EqDefined
  | MoreDefined
  | Unrelated
  deriving Eq

compareExperiments :: TestExperiment -> TestExperiment -> DefinedComparison
compareExperiments te1 te2 = compareRoots (teRoot te1) (teRoot te2) `mergeD` compareArgs (teArgs te1) (teArgs te2)

compareRoots (TEIdRoot x) (TEIdRoot y)
  | x == y    = EqDefined
compareRoots (TEDestructRoot exp1 ctr1 a1 i1 idx1) (TEDestructRoot exp2 ctr2 a2 i2 idx2)
  | ctr1 == ctr2 && idx1 == idx2 && a1 == a2 && i1 == i2 = compareExperiments exp1 exp2
compareRoots (TEConstantRoot c1) (TEConstantRoot c2)
  | c1 == c2  = EqDefined
compareRoots _ _ = Unrelated

compareArgs args1 args2
  | Seq.length args1 == Seq.length args2 = F.foldr mergeD EqDefined . Seq.zipWith compareArg args1 $ args2
  | otherwise                            = Unrelated

compareArg (TEArgExperiment e1) (TEArgExperiment e2) = compareExperiments e1 e2
compareArg TEArgUnconstrained TEArgUnconstrained     = EqDefined
compareArg (TEArgBottom _) (TEArgBottom _)           = EqDefined
compareArg (TEArgBottom _) _                         = LessDefined
compareArg _               (TEArgBottom _)           = MoreDefined
compareArg _               _                         = Unrelated

LessDefined `mergeD` LessDefined = LessDefined
LessDefined `mergeD` EqDefined   = LessDefined
LessDefined `mergeD` _           = Unrelated
EqDefined   `mergeD` x           = x
MoreDefined `mergeD` MoreDefined = MoreDefined
MoreDefined `mergeD` EqDefined   = MoreDefined
MoreDefined `mergeD` _           = Unrelated
Unrelated   `mergeD` _           = Unrelated

mkTestCase :: E.Experiment -> E.ExperimentResult -> Maybe TestCase
mkTestCase (fst . E.canonicaliseExperiment -> exp) expRes = do
  tr <- mkTestResult expRes
  let exp' = mkTestExperiment exp
  return $ TestCase { testExperiment = exp'
                    , testResult     = tr
                    }

mkTestExperiment :: E.Experiment -> TestExperiment
mkTestExperiment e
  = TestExperiment { teRoot = mkTestExperimentRoot (E.experimentRoot e)
                   , teArgs = mkTestExperimentArgs (E.experimentArgs e)
                   }

mkTestResult :: E.ExperimentResult -> Maybe TestResult
mkTestResult (E.Redundant _) = Nothing
mkTestResult E.Ok            = Nothing
mkTestResult (E.Errors e)    = Just $ Errors (fromString . show $ e)
mkTestResult (E.Bottoms b)   = Nothing
mkTestResult (E.Property b)  = Just $ Property b
mkTestResult (E.Show s)      = Just $ Show (fromString s)

mkTestExperimentArgs :: Seq E.ExperimentArg -> Seq TestExperimentArg
mkTestExperimentArgs = fmap fromJust . Seq.filter isJust . fmap mkTestExperimentArg

mkTestExperimentArg :: E.ExperimentArg -> Maybe TestExperimentArg
mkTestExperimentArg (E.ArgExperiment exp) = Just $ TEArgExperiment (mkTestExperiment exp)
mkTestExperimentArg E.ArgUnconstrained    = Just $ TEArgUnconstrained
mkTestExperimentArg (E.ArgBottom _ b)     = Just $ TEArgBottom b
mkTestExperimentArg (E.ArgType _)         = Nothing

mkTestExperimentRoot :: E.ExperimentRoot -> TestExperimentRoot
mkTestExperimentRoot (E.IdRoot id)                = TEIdRoot (fromString . showSDocUnqual . ppr $ id)
mkTestExperimentRoot (E.DestructRoot e dc idx _)  = TEDestructRoot (mkTestExperiment e)
                                                                   (fromString . showSDocUnqual . ppr $ dc)
                                                                   (length $ dataConOrigArgTys dc)
                                                                   (dataConIsInfix dc)
                                                                   idx
mkTestExperimentRoot (E.ConstantRoot c)           = TEConstantRoot (mkConstant c)

mkConstant (E.IntConstant     x)   = TECInt x
mkConstant (E.CharConstant    x)   = TECChar x
mkConstant (E.FloatConstant   x)   = TECFloat x
mkConstant (E.DoubleConstant  x)   = TECDouble x
mkConstant (E.IntegerConstant x _) = TECInteger x

buildComparison :: Map TestExperiment TestResult -> Map TestExperiment TestResult -> Map TestExperiment TestComparison
buildComparison aSuite bSuite = loop init bCases Map.empty
  where
    init = Map.map mkInitialComparison aSuite
    bCases = map (uncurry TestCase) . Map.toList $ bSuite

    loop :: Map TestExperiment TestComparison -> [TestCase] -> Map TestExperiment TestComparison -> Map TestExperiment TestComparison
    loop !active []       !partial  = Map.unionWith (error "duplicate key in buildComparison:loop!") partial active
    loop !active (tc:tcs) !partial  = loop newActive tcs (maybe partial (\(k,v) -> Map.insertWith (error "duplicate key II") k v partial) mPartial)
      where
        (newActive, mPartial) = addTestCase active tc
                                        

    mkInitialComparison :: TestResult -> TestComparison
    mkInitialComparison tr = TC { moreGeneral = A
                                , general     = tr
                                , specific    = Set.empty
                                }

addTestCase :: Map TestExperiment TestComparison -> TestCase -> (Map TestExperiment TestComparison, Maybe (TestExperiment, TestComparison))
addTestCase tes tc
  | Just v <- Map.lookup (testExperiment tc) tes = (Map.delete (testExperiment tc) tes      , Just (testExperiment tc, insertTc v))
  | tcIsLessDefined                              = (unRelated                               , Just (testExperiment tc, lessDefinedComparsion))
  | otherwise                                    = (Map.adjust insertTc tcMoreDefinedExp tes, Nothing)
  where
    (unRelated, related) = Map.partitionWithKey (\k _ -> Unrelated == compareExperiments (testExperiment tc) k) tes

    tcIsLessDefined       = all ((==) LessDefined . (testExperiment tc `compareExperiments`) . fst) $ Map.toList related
    lessDefinedComparsion = TC { moreGeneral = B
                               , general     = testResult tc
                               , specific    = Set.fromList . map (uncurry extractTestCase) . Map.toList $ related
                               }

    tcMoreDefinedExp
      | [(k,v)] <- Map.toList related
      , MoreDefined <- (testExperiment tc) `compareExperiments` k = k
      | otherwise = error "Invariant failure (only one expression can be more defined than this) in addTestCase"
    
    insertTc :: TestComparison -> TestComparison
    insertTc tcmp = tcmp { specific = Set.insert tc . specific $ tcmp }


    extractTestCase :: TestExperiment -> TestComparison -> TestCase
    extractTestCase te (TC { moreGeneral = A
                           , general     = res
                           , specific    = s
                           })
      | Set.null s = TestCase { testExperiment = te
                              , testResult     = res
                              }
      | otherwise = error "Invariant error in extractTestCase"

